/**
 * $Id$
 *
 * Copyright (C) shinichiro.h <s31552@mail.ecc.u-tokyo.ac.jp>
 *  http://user.ecc.u-tokyo.ac.jp/~s31552/wp/
 */

#ifndef approxmath_h_
#define approxmath_h_

#ifndef APPROXMATH_NUMTYPE
#define APPROXMATH_NUMTYPE float
#endif

#define APPROXMATH_HAVE_INLINE
#ifdef APPROXMATH_HAVE_INLINE
#define APPROXMATH_INLINE inline
#else
#define APPROXMATH_INLINE
#endif

#define APPROXMATH_SIN_DOUBLE_ANGLE(suf, x) do { \
  APPROXMATH_NUMTYPE  DA_v = x * 0.5; \
  return 2 * sin##suf(DA_v) * cos##suf(DA_v); \
} while(0)

#define DEFINE_SIN_DOUBLE_ANGLE_FUNC(suf1, suf2) \
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin##suf1(APPROXMATH_NUMTYPE x) { \
    APPROXMATH_SIN_DOUBLE_ANGLE(suf2, x); \
}

#define APPROXMATH_COS_S_DOUBLE_ANGLE(suf, x) do { \
  APPROXMATH_NUMTYPE  DA_v = sin##suf(x * 0.5); \
  return 1 - 2 * DA_v * DA_v; \
} while(0)
#define APPROXMATH_COS_C_DOUBLE_ANGLE(suf, x) do { \
  APPROXMATH_NUMTYPE  DA_v = cos##suf(x * 0.5); \
  return 2 * DA_v * DA_v - 1; \
} while(0)

#define DEFINE_COS_S_DOUBLE_ANGLE_FUNC(suf1, suf2) \
APPROXMATH_INLINE APPROXMATH_NUMTYPE cos##suf1(APPROXMATH_NUMTYPE x) { \
    APPROXMATH_COS_S_DOUBLE_ANGLE(suf2, x); \
}
#define DEFINE_COS_C_DOUBLE_ANGLE_FUNC(suf1, suf2) \
APPROXMATH_INLINE APPROXMATH_NUMTYPE cos##suf1(APPROXMATH_NUMTYPE x) { \
    APPROXMATH_COS_C_DOUBLE_ANGLE(suf2, x); \
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin1(APPROXMATH_NUMTYPE x) {
    return x;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos0(APPROXMATH_NUMTYPE x) {
    return 1;
}

/*
 * _p2 means pi/2. valid range is [-pi/2:pi/2]
 */

#define sin0_p2 sin1
#define sin1_p2 sin1
#define sin2_p2 sin1
#define sin4_p2 sin3_p2
#define sin6_p2 sin5_p2
#define cos0_p2 cos0
#define cos1_p2 cos0
#define cos3_p2 cos2_p2
#define cos5_p2 cos4_p2
#define cos7_p2 cos6_p2

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin3_p2(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    return 0.988792*x-0.145062*x3;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin5_p2(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    APPROXMATH_NUMTYPE x5 = x3*x*x;
    return 0.999771*x-0.165827*x3+0.00757425*x5;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin7_p2(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    APPROXMATH_NUMTYPE x5 = x3*x*x;
    APPROXMATH_NUMTYPE x7 = x5*x*x;
    return 0.999997*x-0.166652*x3+0.00830952*x5-0.000184472*x7;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos2_p2(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x2 = x*x;
    return 0.980163-0.417698*x2;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos4_p2(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x2 = x*x;
    APPROXMATH_NUMTYPE x4 = x2*x2;
    return 0.99958-0.496392*x2+0.0372093*x4;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos6_p2(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x2 = x*x;
    APPROXMATH_NUMTYPE x4 = x2*x2;
    APPROXMATH_NUMTYPE x6 = x4*x2;
    return 0.999995-0.499931*x2+0.0415117*x4-0.00127871*x6;
}

/*
 * _p4 means pi/4. valid range is [-pi/4:pi/4]
 */

#define sin0_p4 sin1
#define sin1_p4 sin1
#define sin2_p4 sin1
#define sin4_p4 sin3_p4
#define sin6_p4 sin5_p4
#define cos0_p4 cos0
#define cos1_p4 cos0
#define cos3_p4 cos2_p4
#define cos5_p4 cos4_p4
#define cos7_p4 cos6_p4

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin3_p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    return 0.999259*x-0.161035*x3;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin5_p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    APPROXMATH_NUMTYPE x5 = x3*x*x;
    return 0.999996*x-0.166612*x3+0.00813765*x5;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE sin7_p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    APPROXMATH_NUMTYPE x5 = x3*x*x;
    APPROXMATH_NUMTYPE x7 = x5*x*x;
    return x-0.166666*x3+0.0083318*x5-0.000194842*x7;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos2_p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x2 = x*x;
    return 0.998672-0.478344*x2;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos4_p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x2 = x*x;
    APPROXMATH_NUMTYPE x4 = x2*x2;
    return 0.999993-0.499764*x2+0.0405122*x4;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE cos6_p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x2 = x*x;
    APPROXMATH_NUMTYPE x4 = x2*x2;
    APPROXMATH_NUMTYPE x6 = x4*x2;
    return 1-0.499999*x2+0.0416566*x4-0.00136059*x6;
}

/*
 * _2p4 means 2*pi/4. valid range is [-pi/2:pi/2]
 * using double angle formula.
 */

DEFINE_SIN_DOUBLE_ANGLE_FUNC(0_2p4, 0_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(1_2p4, 1_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(2_2p4, 2_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(3_2p4, 3_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(4_2p4, 4_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(5_2p4, 5_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(6_2p4, 6_p4);
DEFINE_SIN_DOUBLE_ANGLE_FUNC(7_2p4, 7_p4);

DEFINE_COS_C_DOUBLE_ANGLE_FUNC(0_2p4, 0_p4);
DEFINE_COS_S_DOUBLE_ANGLE_FUNC(1_2p4, 1_p4);
DEFINE_COS_C_DOUBLE_ANGLE_FUNC(2_2p4, 2_p4);
DEFINE_COS_S_DOUBLE_ANGLE_FUNC(3_2p4, 3_p4);
DEFINE_COS_C_DOUBLE_ANGLE_FUNC(4_2p4, 4_p4);
DEFINE_COS_S_DOUBLE_ANGLE_FUNC(5_2p4, 5_p4);
DEFINE_COS_C_DOUBLE_ANGLE_FUNC(6_2p4, 6_p4);
DEFINE_COS_S_DOUBLE_ANGLE_FUNC(7_2p4, 7_p4);

/*
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin3_2p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_SIN_DOUBLE_ANGLE(3_p4, x);
}
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin5_2p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_SIN_DOUBLE_ANGLE(5_p4, x);
}
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin7_2p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_SIN_DOUBLE_ANGLE(7_p4, x);
}
*/

/*
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin3_2p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_COS_DOUBLE_ANGLE(3_p4, x);
}
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin5_2p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_COS_DOUBLE_ANGLE(5_p4, x);
}
APPROXMATH_INLINE APPROXMATH_NUMTYPE sin7_2p4(APPROXMATH_NUMTYPE x) {
    APPROXMATH_COS_DOUBLE_ANGLE(7_p4, x);
}
*/

APPROXMATH_INLINE APPROXMATH_NUMTYPE taylor_sin3(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    return x-x3/6;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE taylor_sin5(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    APPROXMATH_NUMTYPE x5 = x3*x*x;
    return x-x3/6+x5/120;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE taylor_sin7(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE x3 = x*x*x;
    APPROXMATH_NUMTYPE x5 = x3*x*x;
    APPROXMATH_NUMTYPE x7 = x5*x*x;
    return x-x3/6+x5/120-x7/5040;
}

APPROXMATH_INLINE APPROXMATH_NUMTYPE taylor_sininf(APPROXMATH_NUMTYPE x) {
    APPROXMATH_NUMTYPE v = x;
    APPROXMATH_NUMTYPE s = x;
    APPROXMATH_NUMTYPE pv = v;
    int d = 2;
    while (1) {
        s *= x*x;
        s /= d;
        d++;
        s /= d;
        d++;
        if ((d & 2) == 2) v += s;
        else v -= s;
        if (pv == v) break;
        pv = v;
    }
    return v;
}

#endif // ! approxmath_h_

