#include <math.h>
#include "approxmath.h"

typedef struct {
    double sum;
    double avg;
    double max;
    float maxpnt;
} Err;

Err calc_error(float (*orig) (float), float (*approx) (float),
                  float start, float end, int divs)
{
    float d = (end-start)/divs;
    int i;
    float v = start;
    Err err;
    err.sum = 0;
    err.max = 0;
    for (i = 0; i < divs; i++) {
        float r1 = (*orig)(v);
        float r2 = (*approx)(v);
        float rd = fabs(r1-r2);
        err.sum += rd;
        if (err.max < rd) {
            err.max = rd;
            err.maxpnt = v;
        }
        v += d;
    }
    err.avg = err.sum / divs;
    return err;
}

#define SHOW(m, v) do { \
  Err SHOW_v = v; \
  printf("%s:\t %e\t%e\t%f\n", m, SHOW_v.avg, SHOW_v.max, SHOW_v.maxpnt); \
} while (0)

#define SHOW_ERR(m, of, af) SHOW(m, calc_error(&of, &af, s, e, d))

int main() {
    float s, e, d;
    s = -M_PI / 2;
    e = M_PI / 2;
    d = 1000000;

    printf("range is from %f to %f\n", s, e);

    printf("test name:\t AVG of ERR\tMAX of ERR\twhen MAX?\n");

    SHOW_ERR("sin1_p2", sinf, sin1_p2);
    SHOW_ERR("sin3_p2", sinf, sin3_p2);
    SHOW_ERR("sin5_p2", sinf, sin5_p2);
    SHOW_ERR("sin7_p2", sinf, sin7_p2);

    SHOW_ERR("sin1_p4(W)", sinf, sin1_p4);
    SHOW_ERR("sin3_p4(W)", sinf, sin3_p4);
    SHOW_ERR("sin5_p4(W)", sinf, sin5_p4);
    SHOW_ERR("sin7_p4(W)", sinf, sin7_p4);

    SHOW_ERR("cos0_p2", cosf, cos0_p2);
    SHOW_ERR("cos2_p2", cosf, cos2_p2);
    SHOW_ERR("cos4_p2", cosf, cos4_p2);
    SHOW_ERR("cos6_p2", cosf, cos6_p2);

    SHOW_ERR("cos0_p4(W)", cosf, cos0_p4);
    SHOW_ERR("cos2_p4(W)", cosf, cos2_p4);
    SHOW_ERR("cos4_p4(W)", cosf, cos4_p4);
    SHOW_ERR("cos6_p4(W)", cosf, cos6_p4);

    SHOW_ERR("sin1_2p4", sinf, sin1_2p4);
    SHOW_ERR("sin2_2p4", sinf, sin2_2p4);
    SHOW_ERR("sin3_2p4", sinf, sin3_2p4);
    SHOW_ERR("sin4_2p4", sinf, sin4_2p4);
    SHOW_ERR("sin5_2p4", sinf, sin5_2p4);
    SHOW_ERR("sin6_2p4", sinf, sin6_2p4);
    SHOW_ERR("sin7_2p4", sinf, sin7_2p4);

    SHOW_ERR("cos1_2p4", cosf, cos1_2p4);
    SHOW_ERR("cos2_2p4", cosf, cos2_2p4);
    SHOW_ERR("cos3_2p4", cosf, cos3_2p4);
    SHOW_ERR("cos4_2p4", cosf, cos4_2p4);
    SHOW_ERR("cos5_2p4", cosf, cos5_2p4);
    SHOW_ERR("cos6_2p4", cosf, cos6_2p4);
    SHOW_ERR("cos7_2p4", cosf, cos7_2p4);

    SHOW_ERR("taylor sin3", sinf, taylor_sin3);
    SHOW_ERR("taylor sin5", sinf, taylor_sin5);
    SHOW_ERR("taylor sin7", sinf, taylor_sin7);

    SHOW_ERR("taylor sininf", sinf, taylor_sininf);

    s = -M_PI / 4;
    e = M_PI / 4;

    printf("\n");

    printf("range is from %f to %f\n", s, e);

    printf("test name:\t AVG of ERR\tMAX of ERR\twhen MAX?\n");

    SHOW_ERR("sin1_p2", sinf, sin1_p2);
    SHOW_ERR("sin3_p2", sinf, sin3_p2);
    SHOW_ERR("sin5_p2", sinf, sin5_p2);
    SHOW_ERR("sin7_p2", sinf, sin7_p2);

    SHOW_ERR("sin1_p4", sinf, sin1_p4);
    SHOW_ERR("sin3_p4", sinf, sin3_p4);
    SHOW_ERR("sin5_p4", sinf, sin5_p4);
    SHOW_ERR("sin7_p4", sinf, sin7_p4);

    SHOW_ERR("taylor sin3", sinf, taylor_sin3);
    SHOW_ERR("taylor sin5", sinf, taylor_sin5);
    SHOW_ERR("taylor sin7", sinf, taylor_sin7);

    SHOW_ERR("taylor sininf", sinf, taylor_sininf);

    return 0;
}
